/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.dhcp.soap
 * File:    SOAPMessageSender.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.dhcp.soap;

import java.rmi.RemoteException;

import java.util.ArrayList;
import java.util.GregorianCalendar;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.axis2.AxisFault;

import org.trustedcomputinggroup._2010.ifmap._2.DeleteType;
import org.trustedcomputinggroup._2010.ifmap._2.EndSessionType;
import org.trustedcomputinggroup._2010.ifmap._2.IPAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.MACAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.MetadataListType;
import org.trustedcomputinggroup._2010.ifmap._2.NewSessionRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.PublishRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.PurgePublisherRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.RenewSessionType;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType;
import org.trustedcomputinggroup._2010.ifmap._2.UpdateType;
import org.trustedcomputinggroup._2010.ifmap_metadata._2.IpMac;

//import de.fhhannover.inform.iron.dhcp.control.LogWriter;
import de.fhhannover.inform.iron.dhcp.control.Config;

/**
 * This Class contains all methods that are necessary for the communication with the IF-MAP Server
 *
 * @version 	0.01 01 Dez 2009
 * @version 	0.02 08 Dez 2009
 * @version 	0.03 15 Dez 2009
 * @version     0.04 17 Dez 2009
 * @version 	0.05 21 Dez 2009
 * @version 	0.06 22 Dez 2009
 * @version		0.07 17 Mar 2010
 * @version		0.08 30 Mar 2010
 * @version		0.09 13 Apr 2010
 * @version		0.10 13 Apr 2010
 * @version		0.11 20 Apr 2010
 * @version		0.12 27 Apr 2010
 * @version		0.13 04 May 2010
 * @version		0.14 11 May 2010
 * @version		0.15 18 May 2010
 * @version		0.16 22 Jun 2010
 * @author 	Sebastian Kobert
 */
public final class SOAPMessageSender {
	
	/** mSessionID represents the Session ID with the IF-MAP Server */
	private String mSessionID = null;
	
	/** mIfmapPublisherID represents the Publisher ID of the IF-MAP Client */
	private String mIfmapPublisherID = null;
	
	/** mConnection represents the HTTPS Connection with the IF-MAP Server */
	private SecureIFMap mConnection = null;
	
	/** mUpdateMessages represents the Lease Information that should be send to the IF-MAP Server */
	private ArrayList<LeaseMessage> mMessages = null;
	
	/** get a Factory that can Create the Request Types */
	private org.trustedcomputinggroup._2010.ifmap._2.ObjectFactory mBaseFactory = new org.trustedcomputinggroup._2010.ifmap._2.ObjectFactory();
	
	/** get a Factory that can Create the Metadata Types */
	private org.trustedcomputinggroup._2010.ifmap_metadata._2.ObjectFactory mMetadataFactory = new org.trustedcomputinggroup._2010.ifmap_metadata._2.ObjectFactory();
		
    /** 
     * mInstance is for Singleton 
     */
    private static SOAPMessageSender mInstance;

    /**
     * worker_renew a Thread which sends periodically a renewSession()
     */
    private Thread worker_renew = new Thread(){    	
    	public void run(){     		
    		while(!done){
    			if(renew){
    				
					if(!SOAPMessageSender.this.renewSession()){
						renew = false;
					}
    			}    		
    			try {
					Thread.sleep(20000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}    		
    	  }
    	}    
    };
   
    /**
     * renew the renewSession()-Thread should send requests
     * done the renewSession()-Thread will be stopped
     */
    private boolean renew = false;
    private boolean done = false;
    
    /** This is the private Constructor for the Singleton Pattern*/
    private SOAPMessageSender() {}

    /** 
     * static method "getInstance()" delivers the only Instance of this Class
     * it is synchronized (thread-safe)
     */
    public synchronized static SOAPMessageSender getInstance() {
        if (mInstance == null) {
            mInstance = new SOAPMessageSender();
        }
        return mInstance;
    }
    
   
    /**
     * starts the renewSession()-Thread
     */
    public void activateReNewSession(boolean action){
    	if(!worker_renew.isAlive()){
    		if(action){
    			worker_renew.start();
    		}
    	}
    	else{
    		if(!action){
    			done = true;
    		}
    	}
    		
    }
    
	/**
	 * This Method controls the publishing of the Lease Information to the IF-MAP Server 
	 * 
	 * @param messages The ArrayList of LeaseMessages that should be published 
	 */
	public void sendMessages(ArrayList<LeaseMessage> messages){
			renew = false;
			if(messages != null){
				//if there are no messages already, add the given messages
				if(this.mMessages == null || this.mMessages.size() == 0){
					this.mMessages = messages;
				}
				//if there are already messages, add the new messages to the others
				else {				
					for(int i=0; i<messages.size(); i++){
						this.mMessages.add(messages.get(i));					
					}
				}
			}
			
			//################################### connection ########################################################################
			if(this.mConnection == null){
				
				//Open a new connection with the IF-MAP Server
				try {
					this.newConnection();
				} catch (AxisFault e) {
					e.printStackTrace();
					this.mConnection = null;
					return;
				}
			}

			//################################### newSession ########################################################################
			
			if(this.mSessionID == null){

				//create a new Session
				System.out.print("newSession(): ");					

				if(this.newSession() == false){
					return;
				} 					
					

			}
			
			
			//################################### publishUpdate or publishDelete ########################################################################
			int i = 0;
			int j = 0;
			
			while(!mMessages.isEmpty()){
				try {
					LeaseMessage message = mMessages.get(0);
					
					if(message.getMessageType() == LeaseMessage.UPDATE_MESSAGE){
						
						String msg = "["+i+"]"+"PublishUpdate - "+message.getIp()+ " - ";
						System.out.println(msg);
					//	LogWriter.log(msg);						
	
						if(this.publishUpdate(message) == true){
							mMessages.remove(message);
						} else {
							return;
						}
						i++;						
					} else {					
						
						String msg = "["+j+"]"+"PublishDelete - "+message.getIp()+ " - ";
						System.out.println(msg);
					//	LogWriter.log(msg);						
																
						if(this.publishDelete(message) == true){
							mMessages.remove(message);
						} else {
							return;
						}
						j++;
					}
				
				} catch (RemoteException e) {
					e.printStackTrace();
					mConnection = null;
					mSessionID = null;
					return;
				}
			}
			renew = true;
	}
	
	/**
	 * This Method opens a new Connection to the IF-MAP Server 
	 * 
	 * @throws AxisFault 
	 */
	private void newConnection() throws AxisFault{
		try{
			mConnection = new SecureIFMap(Config.MAPSERVER_URL, //URL to Webserver
					Config.MAPSERVER_KEYSTORE_PATH,  //Keystore for the public key
					Config.MAPSERVER_TRUSTSTORE_PATH); //Keystore for the private key
		} catch (AxisFault a){
			a.printStackTrace();
			mConnection = null;
			throw a;
		}
		
	}
	
	/**
	 * This Method opens a new Session with the IF-MAP Server 
	 * 
	 * @return true - if newSession was successful, false - otherwise
	 * @throws RemoteException 
	 */
	private boolean newSession() {
			
		ResponseType response = null;
    
		NewSessionRequestType newSessioRequestType = mBaseFactory.createNewSessionRequestType();
	
		try{				
            if(mConnection == null){
            	return false;
            } else {
            	response = mConnection.newSession(newSessioRequestType);
            }
		} catch(RemoteException e){
			e.printStackTrace();
			mConnection = null;
			mSessionID = null;
			return false;
		}

		if(response.getErrorResult() == null){
			System.out.println("Publisher-ID: "+response.getNewSessionResult().getIfmapPublisherId());
			System.out.println("Session-ID: " +response.getNewSessionResult().getSessionId());
			this.mSessionID = response.getNewSessionResult().getSessionId();
			this.mIfmapPublisherID = response.getNewSessionResult().getIfmapPublisherId();
			return true;
		} else {
			this.printError("newSession", response);
			mSessionID = null;
			return false;
		}	
	}
    
	/**
	 * This Method renews the Session with the IF-MAP Server 
	 * 
	 * @return true - if renewSession was successful, false - otherwise  		
	 * @throws RemoteException 
	 */
    private boolean renewSession() {
    	
    	ResponseType response = null;
	
		RenewSessionType renewSessionRequestType = mBaseFactory.createRenewSessionType();
			
		renewSessionRequestType.setSessionId(mSessionID);

		try{
            if(mConnection == null){
            	return false;
            } else {
            	response = mConnection.renewSession(renewSessionRequestType);
            }
		} catch(RemoteException e){
			e.printStackTrace();
			mConnection = null;
			mSessionID = null;
			return false;
		}
		
		
		if(response.getErrorResult() == null){			
			return true;
		} else {
			this.printError("renewSession", response);
			mSessionID = null;
			return false;
		}
    }
    
	/**
	 * This Method ends the Session with the IF-MAP Server 
	 * 
	 * @return true - if endSession was successful, false - otherwise 
	 * @throws RemoteException 
	 */
    private boolean endSession() {
	
    	ResponseType response = null;
    	
        EndSessionType endSessionRequestType = mBaseFactory.createEndSessionType();
        	
        endSessionRequestType.setSessionId(mSessionID);
        
        try{
            if(mConnection == null){
            	return false;
            } else {
            	response = mConnection.endSession(endSessionRequestType);
            }
		} catch(RemoteException e){
			e.printStackTrace();
			mConnection = null;
			mSessionID = null;
			return false;
		}
		
		if(response.getErrorResult() == null){			
			mSessionID = null;			
			return true;
		} else {
			this.printError("endSession", response);
			return false;
		}
    }
    
	/**
	 * This Method sends a purgePublisher Request to the IF-MAP Server 
	 * 
	 * @return true - if purgePublish was successful, false - otherwise 
	 * @throws RemoteException 
	 */
    private boolean purgePublish() {
    	
    	ResponseType response;
    	
    	PurgePublisherRequestType newPurgePublish = mBaseFactory.createPurgePublisherRequestType();
    	
    	newPurgePublish.setIfmapPublisherId(mIfmapPublisherID);
    	
    	newPurgePublish.setSessionId(mSessionID);
	
    	try{
            if(mConnection == null){
            	return false;
            } else {
            	response = mConnection.purgePublisher(newPurgePublish);
            }
		} catch(RemoteException e){
			e.printStackTrace();
			mConnection = null;
			mSessionID = null;
			return false;
		}
		
		if(response.getErrorResult() != null){
			this.printError("purgePublish", response);
			mSessionID = null;
			return false;
		}
		
		return true;
    }
    
	/**
	 * This Method publishes the Lease Information to the IF-MAP Server 
	 * 
	 * @param message The LeaseMessage that should be published
	 * @return true - if publishUpdate was successful, false - otherwise
	 * @throws RemoteException 
	 */
    private boolean publishUpdate(LeaseMessage message) throws RemoteException{
    	
    		//Create Metadata Object
    		IpMac ipMac = mMetadataFactory.createIpMac();
    		
    		//put Data into the Metadata Object
    		//set the IP Adress of the DHCP Server
    		ipMac.setDhcpServer(Config.DHCPD_SERVER_IP);		
    		
    		try {
    			GregorianCalendar gc = new GregorianCalendar();
    			gc.setTime(message.getStartDate().getTime());
    			
    			//set the Start Time of the DHCP Lease
				ipMac.setStartTime(DatatypeFactory.newInstance().newXMLGregorianCalendar(gc));				
				
				gc = new GregorianCalendar();
				gc.setTime(message.getEndDate().getTime());
				
				//set the End Time of the DHCP Lease
				ipMac.setEndTime(DatatypeFactory.newInstance().newXMLGregorianCalendar(gc));
				
			} catch (DatatypeConfigurationException e) {				
				e.printStackTrace();
				return false;
			}
			
    		ipMac.setIfmapCardinality("multiValue");    		
    		
    		//###########################################################
    		
    		//create a Container for the Metadata
    		MetadataListType metadataListType = mBaseFactory.createMetadataListType();
    		//add the Metadata Object to the Container
    		metadataListType.getAny().add(ipMac);
    		
    		//###########################################################
    		
    		//set the IP Address of the DHCP Client
    		IPAddressType ipAddressType = mBaseFactory.createIPAddressType();
    		
    		ipAddressType.setValue(message.getIp());
    		ipAddressType.setType("IPv4");		
    		
    		//###########################################################
    		
    		//set the MAC Address of the DHCP Client
    		MACAddressType macAddressType = mBaseFactory.createMACAddressType();
    		
    		macAddressType.setValue(message.getMac());
    		
    		//###########################################################
    		
    		UpdateType updateType = mBaseFactory.createUpdateType();
    		
    		updateType.setLifetime("forever");		
    		updateType.getAccessRequestOrIdentityOrIpAddress().add(ipAddressType);
    		updateType.getAccessRequestOrIdentityOrIpAddress().add(macAddressType);
    		updateType.setMetadata(metadataListType);
    		
    		//###########################################################    		
    		
    		PublishRequestType newPublish = mBaseFactory.createPublishRequestType();		
    		
    		newPublish.setSessionId(mSessionID);		
    		newPublish.getUpdateOrNotifyOrDelete().add(updateType);
    		
    		//###########################################################
    		
    		ResponseType response = null;
    		
    		response = mConnection.publish(newPublish);   		
						
			if(response.getErrorResult() != null){				
				this.printError("publishUpdate", response);
				mSessionID = null;
				return false;
			}
			
			return true;
    }
    
    	/**
    	 * This Method deletes published Lease Information from the IF-MAP Server
    	 * 
    	 * @param message The LeaseMessage that should be deleted
    	 * @return true - if publishDelete was successful, false - otherwise 
    	 * @throws RemoteException 
    	 */
    private boolean publishDelete(LeaseMessage message) throws RemoteException{    	
			
    		XMLGregorianCalendar xmlGcStartTime = null;
    		XMLGregorianCalendar xmlGcEndTime = null;
    		String filterString;
    	
    		//get the right Data Format for the Start- and End-Date, so it can be used in the delete String
			try {
				GregorianCalendar gc = new GregorianCalendar();
				gc.setTime(message.getStartDate().getTime());
				xmlGcStartTime = DatatypeFactory.newInstance().newXMLGregorianCalendar(gc);
				
				gc = new GregorianCalendar();
				gc.setTime(message.getEndDate().getTime());
				xmlGcEndTime = DatatypeFactory.newInstance().newXMLGregorianCalendar(gc);
				
				filterString = "ns3:ip-mac[@ifmap-publisher-id=\"" + this.mIfmapPublisherID + "\"]";
				
			} catch (DatatypeConfigurationException e) {
				e.printStackTrace();				
				return false;
			}	
	
			//###########################################################
	
			//set the IP Address of the DHCP Client
    		IPAddressType ipAddressType = mBaseFactory.createIPAddressType();
    		
    		ipAddressType.setValue(message.getIp());
    		ipAddressType.setType("IPv4");		
    		
    		//###########################################################
    		
    		//set the MAC Address of the DHCP Client
    		MACAddressType macAddressType = mBaseFactory.createMACAddressType();
    		
    		macAddressType.setValue(message.getMac());
    		
    		//###########################################################
    		
    		DeleteType deleteType = mBaseFactory.createDeleteType();
    		deleteType.getAccessRequestOrIdentityOrIpAddress().add(ipAddressType);
    		deleteType.getAccessRequestOrIdentityOrIpAddress().add(macAddressType);
    		deleteType.setFilter(filterString);
    		
    		//###########################################################
    		
    		PublishRequestType newPublish = mBaseFactory.createPublishRequestType();		
    		
    		newPublish.setSessionId(mSessionID);		
    		newPublish.getUpdateOrNotifyOrDelete().add(deleteType);
    		
    		//###########################################################
    		
    		ResponseType response = null;
    		
    		response = mConnection.publish(newPublish);
    		
			if(response.getErrorResult() != null){				
				this.printError("publishDelete", response);
				mSessionID = null;
				return false;
			}
			
			return true;
    }
    
	/**
	 * This Method prints a given Error Result on the Error Console
	 * The Error Result is part of the Response Object
	 * 
	 * @param errorLocation In which Method did the Error appear?
	 * @param response The Response Object from the Server
	 */
	private void printError(String errorLocation, ResponseType response) {
		System.err.println("[Error] " + errorLocation + " Response was an Error Result");
		System.err.println("Error String: " + response.getErrorResult().getErrorString());
		System.err.println("Error Code: " + response.getErrorResult().getErrorCode());
		
	}
	
	/**
	 * This Method grants the MapControl the 
	 * 
	 * @return true if purge Publisher was successful, false otherwise
	 */
	public boolean makeAPurgePublish(){
		try {
			this.newConnection();
		} catch (AxisFault e) {
			e.printStackTrace();
			return false;
		}
		
		if(this.newSession() == false){
			return false;
		}
		
		if(this.purgePublish() == false){				
			return false;
		}		

		return true;
	}
	
	/**
	 * This Method allows to call the private Method endSession 
	 * 
	 */
	public void redirectEndSession() {
		this.endSession();
	}
}